# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_cli/_main_.py: CLI auto-detection

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import importlib
import os
import sys

import click

from .. import __version__

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def _add_commands(ctx):
    """auto-detects sub-commands"""
    for cmd in (
        item[:-3] if item.lower().endswith(".py") else item[:]
        for item in os.listdir(os.path.dirname(__file__))
        if not item.startswith("_")
    ):
        try:
            ctx.add_command(
                getattr(importlib.import_module("scherbelberg._cli.%s" % cmd), cmd)
            )
        except ModuleNotFoundError:
            continue


@click.group(invoke_without_command=True)
@click.option("--version", is_flag=True)
def cli(version):
    """HPC cluster deployment and management for the Hetzner Cloud"""

    if not version:
        return

    print(__version__)
    sys.exit()


_add_commands(cli)



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_core/node.py: Cluster nodes (servers)

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from asyncio import sleep
from logging import getLogger, Logger
import os
from typing import Dict, Optional, Union

from hcloud import Client
from hcloud.servers.client import BoundServer

from typeguard import typechecked

from .abc import NodeABC, SSHConfigABC
from .command import Command
from .sshconfig import SSHConfig

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Node(NodeABC):
    """
    Represents one node of the cluster, i.e. a server. Mutable.

    Args:
        server : A cloud API server object.
        client : A cloud API client object.
        fn_private : Location of private SSH key.
        prefix : Name of cluster, used as a prefix in names of every component.
        wait : Timeout in seconds before actions are repeated or exceptions are raised.
        log : Allows to pass custom logger objects. Defaults to scherbelberg's own default logger.
    """

    def __init__(
        self,
        server: BoundServer,
        client: Client,
        fn_private: str,
        prefix: str,
        wait: float,
        log: Union[Logger, None] = None,
    ):

        self._log = getLogger(name=prefix) if log is None else log

        assert len(fn_private) > 0
        assert wait > 0

        self._server = server
        self._client = client
        self._fn_private = fn_private
        self._prefix = prefix
        self._wait = wait

    def __repr__(self) -> str:
        """
        Interactive string representation
        """

        return f"<Node name={self.name:s} public={self.public_ip4:s} private={self.private_ip4}>"

    def _l(self, msg: str) -> str:

        return f"[{self.suffix:s}] {msg:s}"

    async def get_sshconfig(self, user: Optional[str] = None) -> SSHConfigABC:
        """
        Generates SSH configuration for commands that are supposed to be executed on this node.

        Args:
            user : Remote user name, defaults to standard cluster user name.
        Returns:
            SSH configuration.
        """

        if user is None:
            user = f"{self._prefix:s}user"

        return SSHConfig(
            name=self.public_ip4,
            user=user,
            fn_private=self._fn_private,
        )

    async def ping_ssh(self, user: Optional[str] = None) -> bool:
        """
        Checks if node can be contacted via SSH by executing an ``exit`` command on it.
        Performs exactly one single try. Does not raise any type of exception.

        Args:
            user : Remote user name, defaults to standard cluster user name.
        Returns:
            Success (or the lack thereof).
        """

        if user is None:
            user = f"{self._prefix:s}user"

        _, _, status, _ = (
            await Command.from_list(["exit"])
            .on_host(
                host=await self.get_sshconfig(user=user),
            )
            .run(returncode=True, timeout=5, wait=self._wait)
        )

        assert len(status) == 1
        status = status[0]

        return status == 0

    async def reboot(self):
        """
        Triggers a server reboot.
        """

        self._server.reboot()

    async def update(self):
        """
        Updates the internal cloud API server object by requesting new information about the node from the cloud API.
        """

        self._server = self._client.servers.get_by_name(name=self.name)

    async def bootstrap(self):
        """
        Bootstraps scherbelberg's basic infrastructure on the node, i.e.

        - System updates
        - Secure user & SSH configuration
        - TLS/SSL certificate
        - Conda-forge base install via mamba-forge
        """

        await self.wait_for_ssh(user="root")

        self._log.info(self._l("Copying root files to node ..."))
        await Command.from_scp(
            *[
                os.path.abspath(
                    os.path.join(
                        os.path.dirname(__file__),
                        "..",
                        "share",
                        fn,
                    )
                )
                for fn in (
                    "bootstrap_01.sh",
                    "bootstrap_02.sh",
                    "sshd_config.patch",
                )
            ],
            target="~/",
            host=await self.get_sshconfig(user="root"),
        ).run(wait=self._wait)

        self._log.info(self._l("Running first bootstrap script ..."))
        await Command.from_list(["bash", "bootstrap_01.sh"]).on_host(
            host=await self.get_sshconfig(user="root")
        ).run(wait=self._wait)

        self._log.info(self._l("Rebooting ..."))
        await self.reboot()
        await self.wait_for_ssh(user="root")

        self._log.info(self._l("Running second bootstrap script ..."))
        await Command.from_list(["bash", "bootstrap_02.sh", self._prefix]).on_host(
            host=await self.get_sshconfig(user="root")
        ).run(wait=self._wait)
        await self.wait_for_ssh(user=f"{self._prefix:s}user")

        self._log.info(self._l("Copying user files to node ..."))
        await Command.from_scp(
            *[
                os.path.abspath(
                    os.path.join(
                        os.path.dirname(__file__),
                        "..",
                        "share",
                        fn,
                    )
                )
                for fn in (
                    "bootstrap_03.sh",
                    "bootstrap_scheduler.sh",
                    "bootstrap_worker.sh",
                    "requirements_conda.txt",
                    "requirements_pypi.txt",
                )
            ],
            *[
                os.path.abspath(
                    os.path.join(os.getcwd(), f"{self._prefix:s}_{suffix:s}")
                )
                for suffix in (
                    "ca.crt",
                    "node.key",
                    "node.crt",
                )
            ],
            target="~/",
            host=await self.get_sshconfig(),
        ).run(wait=self._wait)

        self._log.info(self._l("Running third (user) bootstrap script ..."))
        await Command.from_list(["bash", "bootstrap_03.sh", self._prefix]).on_host(
            host=await self.get_sshconfig()
        ).run(wait=self._wait)

        self._log.info(self._l("Bootstrapping done."))

    async def start_scheduler(self, dask_ipc: int, dask_dash: int):
        """
        Starts Dask scheduler on node.

        Args:
            dask_ipc : Port used for Dask's interprocess communication.
            dask_dash : Port used for Dask's dashboard.
        """

        assert dask_ipc >= 2 ** 10
        assert dask_dash >= 2 ** 10

        assert dask_ipc != dask_dash

        await self.wait_for_ssh()

        self._log.info(self._l("Staring dask scheduler ..."))

        await Command.from_list(
            [
                "bash",
                "-i",
                "bootstrap_scheduler.sh",
                f"{dask_ipc:d}",
                f"{dask_dash:d}",
                self._prefix,
            ]
        ).on_host(host=await self.get_sshconfig()).run(wait=self._wait)

        self._log.info(self._l("Dask scheduler started."))

    async def start_worker(
        self, dask_ipc: int, dask_dash: int, dask_nanny: int, scheduler_ip4: str
    ):
        """
        Starts Dask scheduler on node.

        Args:
            dask_ipc : Port used for Dask's interprocess communication.
            dask_dash : Port used for Dask's dashboard.
            dask_nanny : Port used for Dask's nanny.
            scheduler_ip4 : IPv4 address of scheduler node.
        """

        assert dask_ipc >= 2 ** 10
        assert dask_dash >= 2 ** 10
        assert dask_nanny >= 2 ** 10

        assert len({dask_ipc, dask_dash, dask_nanny}) == 3

        await self.wait_for_ssh()

        self._log.info(self._l("Staring dask worker ..."))

        await Command.from_list(
            [
                "bash",
                "-i",
                "bootstrap_worker.sh",
                scheduler_ip4,
                f"{dask_ipc:d}",
                f"{dask_dash:d}",
                f"{dask_nanny:d}",
                self._prefix,
            ]
        ).on_host(host=await self.get_sshconfig()).run(wait=self._wait)

        self._log.info(self._l("Dask worker started."))

    async def wait_for_ssh(self, user: Optional[str] = None):
        """
        Keeps pinging the server on SSH via :meth:`scherbelberg.Node.ping_ssh` until success in ``wait`` second intervals.

        Args:
            user : Remote user name, defaults to standard cluster user name.
        """

        if user is None:
            user = f"{self._prefix:s}user"

        self._log.info(self._l("[%s] Waiting for SSH ..."), user)

        while True:
            ssh_up = await self.ping_ssh(user)
            if ssh_up:
                break
            await sleep(self._wait)
            self._log.info(self._l("[%s] Continuing to wait for SSH ..."), user)

        self._log.info(self._l("[%s] SSH up."), user)

    @property
    def name(self) -> str:
        """
        Full name of node / server
        """

        return self._server.name

    @property
    def labels(self) -> Dict[str, str]:
        """
        Labels of node / server
        """

        return self._server.labels.copy()

    @property
    def public_ip4(self) -> str:
        """
        Public IPv4 address of node / server
        """

        return self._server.public_net.ipv4.ip

    @property
    def private_ip4(self) -> str:
        """
        Private IPv4 address of node / server
        """

        assert len(self._server.private_net) == 1

        return self._server.private_net[0].ip

    @property
    def suffix(self) -> str:
        """
        Suffix portion of name of node / server
        """

        return self.name.split("-node-")[1]

    @classmethod
    async def from_async(cls, **kwargs) -> NodeABC:
        """
        Constructs objects from this class asynchronously

        Returns:
            New node object
        """

        return cls(**kwargs)

    @classmethod
    async def from_name(
        cls,
        name: str,
        client: Client,
        fn_private: str,
        prefix: str,
        wait: float,
        log: Union[Logger, None] = None,
    ) -> NodeABC:
        """
        Creates :class:`scherbelberg.Node` object by connecting to an existing server.

        Args:
            name : Full name of node / server.
            client : Cloud API client object.
            fn_private : Location of private SSH key.
            prefix : Name of cluster, used as a prefix in names of every component.
            wait : Timeout in seconds before actions are repeated or exceptions are raised.
            log : Allows to pass custom logger objects. Defaults to scherbelberg's own default logger.
        Returns:
            New node object
        """

        return cls(
            server=client.servers.get_by_name(name=name),
            client=client,
            fn_private=fn_private,
            prefix=prefix,
            wait=wait,
            log=log,
        )



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_core/command.py: Command wrapper

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from asyncio import sleep
import itertools
from subprocess import Popen, PIPE
from typing import List, Tuple, Union
import shlex
from sys import platform
from time import time

from typeguard import typechecked

from .abc import CommandABC, SSHConfigABC
from .const import WAIT
from .process import Process

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Command(CommandABC):
    """
    Representing a chain of commands, connected via pipes. Immutable.

    Args:
        cmd : List of list of strings. Each inner list represents one command compatible to ``subprocess.Popen``.
    """

    def __init__(self, cmd: List[List[str]]):

        self._cmd = [fragment.copy() for fragment in cmd]

    def __repr__(self) -> str:
        """
        Interactive string representation
        """

        return "<Command>"

    def __str__(self) -> str:
        """
        String conversion
        """

        return " | ".join([shlex.join(fragment) for fragment in self._cmd])

    def __len__(self) -> int:
        """
        Number of chained commmands
        """

        return len(self._cmd)

    def __or__(self, other: CommandABC) -> CommandABC:  # pipe
        """
        Pipe
        """

        return type(self)(self.cmd + other.cmd)

    @staticmethod
    def _split_list(data: List, delimiter: str) -> List[List]:

        return [
            list(sub_list)
            for is_delimiter, sub_list in itertools.groupby(
                data, lambda item: item == delimiter
            )
            if not is_delimiter
        ]

    @staticmethod
    def _ssh_options() -> List[str]:

        dev_null = "\\\\.\\NUL" if platform.startswith("win") else "/dev/null"

        return [
            "-o",
            "StrictHostKeyChecking=no",  # TODO security
            "-o",
            f"UserKnownHostsFile={dev_null:s}",  # TODO security
            "-o",
            "BatchMode=yes",
            "-o",
            "ConnectTimeout=5",
        ]

    async def run(
        self,
        returncode: bool = False,
        timeout: Union[float, int, None] = None,
        wait: float = WAIT,
    ) -> Union[
        Tuple[List[str], List[str], List[int], Exception],
        Tuple[List[str], List[str]],
    ]:
        """
        Run command or chain of commands in a :class:`scherbelberg.Process` object.

        Args:
            returncode : If set to ``True``, returns actual return code and does not raise an exception if the command(s) failed. If set to ``False``, a failed command raises an exception and only data from standard output and standard error streams is returned.
            timeout : Total timeout in seconds.
            wait : Interval defining every how many seconds the status of the running process is being observed.
        Returns:
            A tuple, the first two elements containing data from standard output and standard error streams. If ``returncode`` is set to ``True``, the tuple has two additional entries, a list of return codes and an exception object that can be raised by the caller.
        """

        procs = []  # all processes, connected with pipes

        for index, fragment in enumerate(self._cmd):  # create & connect processes

            stdin = None if index == 0 else procs[-1].stdout  # output of last process
            proc = Popen(
                fragment,
                stdout=PIPE,
                stderr=PIPE,
                stdin=stdin,
            )
            procs.append(proc)

        process = Process(
            procs=procs,
            command=self,
        )

        start = time()
        while process.running:
            await sleep(wait)
            if timeout is not None and (time() - start) >= timeout:
                break

        return process.communicate(
            returncode=returncode,
            timeout=0.1 if timeout is not None else None,
        )

    def on_host(self, host: SSHConfigABC) -> CommandABC:
        """
        Adds a ``ssh`` prefix to the command so it can be executed on a remote host.
        Does not change the current command but returns a new one.

        Args:
            host : SSH configuration
        Returns:
            A new :class:`scherbelberg.Command` object which can be executed on the remote host.
        """

        if host.name == "localhost":
            return self

        return type(self).from_list(
            [
                "ssh",
                "-T",  # Disable pseudo-terminal allocation
                "-o",
                "Compression=yes" if host.compression else "Compression=no",
                *self._ssh_options(),
                "-p",
                f"{host.port:d}",
                "-c",
                host.cipher,
                "-i",
                host.fn_private,
                f"{host.user:s}@{host.name:s}",
                str(self),
            ]
        )

    @property
    def cmd(self) -> List[List[str]]:
        """
        List of list of strings. Each inner list represents one command compatible to ``subprocess.Popen``. This interface can not be used to change the command.
        """

        return [fragment.copy() for fragment in self._cmd]

    @classmethod
    def from_str(cls, cmd: str) -> CommandABC:
        """
        Generates a :class:`scherbelberg.Command` object from a single string containing a (shell) command.

        Args:
            cmd : Single string containing a (shell) command.
        Returns:
            New command object.
        """

        return cls(cls._split_list(shlex.split(cmd), "|"))

    @classmethod
    def from_list(cls, cmd: List[str]) -> CommandABC:
        """
        Generates a :class:`scherbelberg.Command` object from a list of strings compatible to ``subprocess.Popen``.

        Args:
            cmd : A list of strings compatible to ``subprocess.Popen``.
        Returns:
            New command object.
        """

        return cls([cmd])

    @classmethod
    def from_scp(cls, *source: str, target: str, host: SSHConfigABC) -> CommandABC:
        """
        Generates a :class:`scherbelberg.Command` object representing an ``scp`` command.
        Only supports copy operations from the local system to the remote host.

        Args:
            source : An arbitrary number of paths on the local system.
            target : Target path on the remote system.
            host : SSH configuration.
        Returns:
            New command object.
        """

        assert len(source) > 0
        assert len(target) > 0

        if platform.startswith("win"):  # Windows scp path fix
            source = [path.replace("\\\\", "/").replace("\\", "/") for path in source]

        return cls.from_list(
            [
                "scp",
                "-o",
                "Compression=yes" if host.compression else "Compression=no",
                *cls._ssh_options(),
                "-P",
                f"{host.port:d}",
                "-c",
                host.cipher,
                "-i",
                host.fn_private,
                *source,
                f"{host.user:s}@{host.name:s}:{target:s}",
            ]
        )



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_core/sshconfig.py: SSH configuration

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Union

from typeguard import typechecked

from .abc import SSHConfigABC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class SSHConfig(SSHConfigABC):
    """
    Holds configuration for an SSH connection. Immutable.

    Args:
        name : Domain name or IP of remote system.
        user : Remote user name.
        fn_private : Location of private SSH key.
        port : SSH port on remote system.
        compression : Turns SSH compression on or off.
        cipher : Specifies cipher for SSH connection.
    """

    def __init__(
        self,
        name: str,
        user: str,
        fn_private: str,
        port: int = 22,
        compression: bool = True,
        cipher: str = "aes256-gcm@openssh.com",
    ):

        assert len(name) > 0
        assert len(user) > 0
        assert len(fn_private) > 0
        assert port > 0
        assert len(cipher) > 0

        self._name = name
        self._user = user
        self._fn_private = fn_private
        self._port = port
        self._compression = compression
        self._cipher = cipher

    def __repr__(self) -> str:
        """
        Interactive string representation
        """

        return f"<SSHConfig {self._user}@{self._name:s}:{self._port:d} compression={'yes' if self._compression else 'no':s} cipher={self._cipher:s}>"

    def new(
        self,
        name: Union[str, None],
        user: Union[str, None],
        fn_private: Union[str, None],
        port: Union[int, None],
        compression: Union[bool, None],
        cipher: Union[str, None],
    ) -> SSHConfigABC:
        """
        Generate a new SSH configuration from present object by changing individual parameters.

        Args:
            name : Domain name or IP of remote system.
            user : Remote user name.
            fn_private : Location of private SSH key.
            port : SSH port on remote system.
            compression : Turns SSH compression on or off.
            cipher : Specifies cipher for SSH connection.
        Returns:
            New SSH configuration object.
        """

        return type(self)(
            name=self._name if name is None else name,
            user=self._user if user is None else user,
            fn_private=self._fn_private if fn_private is None else fn_private,
            port=self._port if port is None else port,
            compression=self._compression if compression is None else compression,
            cipher=self._cipher if cipher is None else cipher,
        )

    @property
    def name(self) -> str:
        """
        Domain name or IP of remote system
        """

        return self._name

    @property
    def user(self) -> str:
        """
        Remote user name
        """

        return self._user

    @property
    def fn_private(self) -> str:
        """
        Location of private SSH key
        """

        return self._fn_private

    @property
    def port(self) -> int:
        """
        SSH port on remote system
        """

        return self._port

    @property
    def compression(self) -> bool:
        """
        Turns SSH compression on or off
        """

        return self._compression

    @property
    def cipher(self) -> str:
        """
        Specifies cipher for SSH connection
        """

        return self._cipher



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_core/process.py: Wrapper for Popen

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from subprocess import Popen, TimeoutExpired
from typing import List, Tuple, Union

from typeguard import typechecked

from .abc import CommandABC, ProcessABC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Process(ProcessABC):
    """
    Wrapper around a list of ``subprocess.Popen`` objects, managing their life-cycle. Mutable.

    Args:
        procs : A list of ``subprocess.Popen`` objects linked via pipes.
        command : The source :class:`scherbelberg.Command` object.
    """

    def __init__(self, procs: List[Popen], command: CommandABC):

        self._procs = procs
        self._command = command

        self._completed = False
        self._output = []
        self._errors = []
        self._status = []
        self._exception = None

    def __repr__(self) -> str:
        """
        Interactive string representation
        """

        return "<Process>"

    def communicate(
        self,
        returncode: bool = False,
        timeout: Union[float, int, None] = None,
    ) -> Union[
        Tuple[List[str], List[str], List[int], Exception],
        Tuple[List[str], List[str]],
    ]:
        """
        Run process or chain of processes.

        Args:
            returncode : If set to ``True``, returns actual return code and does not raise an exception if the process(es) failed. If set to ``False``, a failed process raises an exception and only data from standard output and standard error streams is returned.
            timeout : Total timeout in seconds.
        Returns:
            A tuple, the first two elements containing data from standard output and standard error streams. If ``returncode`` is set to ``True``, the tuple has two additional entries, a list of return codes and an exception object that can be raised by the caller.
        """

        self._complete(timeout=timeout)

        if returncode:
            return self._output, self._errors, self._status, self._exception

        if any((code != 0 for code in self._status)):
            raise self._exception

        return self._output, self._errors

    @property
    def running(self) -> bool:
        """
        Is any of the processes in the list of ``subprocess.Popen`` objects still running?
        """

        return any((proc.poll() is None for proc in self._procs))

    def _complete(
        self,
        timeout: Union[float, int, None] = None,
    ):

        if self._completed:
            return

        for proc in self._procs[::-1]:  # inverse order, last process first

            try:
                out, err = proc.communicate(timeout=timeout)
            except TimeoutExpired:
                proc.kill()
                out, err = proc.communicate()

            self._output.append(self._com_to_str(out))
            self._errors.append(self._com_to_str(err))
            self._status.append(int(proc.returncode))

        self._output.reverse()
        self._errors.reverse()
        self._status.reverse()
        self._exception = SystemError(
            "command failed",
            str(self._command),
            self._output,
            self._errors,
        )

        self._completed = True

    @staticmethod
    def _com_to_str(com: Union[str, bytes, None]) -> str:

        if com is None:
            return ""

        if isinstance(com, bytes):
            try:
                return com.decode("utf-8")
            except UnicodeDecodeError:
                return repr(com)

        return com



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_core/cluster.py: A cluster

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from logging import getLogger, Logger
import os
from typing import Any, List, Union

from hcloud import Client
from hcloud.firewalls.client import BoundFirewall
from hcloud.networks.client import BoundNetwork

from typeguard import typechecked

from .abc import ClusterABC, NodeABC
from .const import (
    DASK_IPC,
    DASK_DASH,
    DASK_NANNY,
    PREFIX,
    TOKENVAR,
    WAIT,
    HETZNER_DATACENTER,
    HETZNER_IMAGE_UBUNTU,
    HETZNER_INSTANCE_TINY,
    WORKERS,
)
from .creator import Creator
from .node import Node

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Cluster(ClusterABC):
    """
    Defines a cluster of nodes. Mutable.
    Use the asynchronous ``from_*`` classmethods to instantiate objects of this class.

    Args:
        client : A cloud-API client object.
        scheduler : A node running the Dask scheduler.
        workers : A list of nodes running Dask workers.
        network : A cloud-API network object.
        firewall : A cloud-API firewall object.
        dask_ipc : Port used for Dask's interprocess communication.
        dask_dash : Port used for Dask's dashboard.
        dask_nanny : Port used for Dask's nanny.
        prefix : Name of cluster, used as a prefix in names of every component.
        wait : Timeout in seconds before actions are repeated or exceptions are raised.
        log : Allows to pass custom logger objects. Defaults to scherbelberg's own default logger.
    """

    def __init__(
        self,
        client: Client,
        scheduler: NodeABC,
        workers: List[NodeABC],
        network: BoundNetwork,
        firewall: BoundFirewall,
        dask_ipc: int = DASK_IPC,
        dask_dash: int = DASK_DASH,
        dask_nanny: int = DASK_NANNY,
        prefix: str = PREFIX,
        wait: float = WAIT,
        log: Union[Logger, None] = None,
    ):

        assert len(workers) > 0

        assert dask_ipc >= 2 ** 10
        assert dask_dash >= 2 ** 10
        assert dask_nanny >= 2 ** 10

        assert len({dask_ipc, dask_dash, dask_nanny}) == 3

        self._client = client
        self._scheduler = scheduler
        self._workers = workers
        self._network = network
        self._firewall = firewall

        self._dask_ipc = dask_ipc  # port
        self._dask_dash = dask_dash  # port
        self._dask_nanny = dask_nanny  # port

        self._prefix = prefix
        self._wait = wait
        self._log = getLogger(name=prefix) if log is None else log

    def __repr__(self) -> str:
        """
        Interactive string representation
        """

        workers = 0 if self._workers is None else len(self._workers)

        return f'<Cluster prefix="{self._prefix:s}" alive={str(self.alive):s} workers={workers:d} ipc={self._dask_ipc:d} dash={self._dask_dash:d} nanny={self._dask_nanny:d}>'

    async def get_client(self, asynchronous: bool = True) -> Any:
        """
        Creates and returns a client.

        Args:
            asynchronous : Specifies if the ``Client`` object runs in asynchronous mode.
        Returns:
            ``dask.distributed.Client`` object attached to the cluster.
        """

        if not self.alive:
            raise SystemError("cluster is dead")

        from dask.distributed import Client as DaskClient
        from distributed.security import Security

        security = Security(
            tls_ca_file=f"{self._prefix:s}_ca.crt",
            tls_client_cert=f"{self._prefix:s}_node.crt",
            tls_client_key=f"{self._prefix:s}_node.key",
            require_encryption=True,
        )

        return DaskClient(
            f"tls://{self._scheduler.public_ip4:s}:{self._dask_ipc:d}",
            security=security,
            asynchronous=asynchronous,
        )

    async def destroy(self):
        """
        Destroys a living cluster
        """

        if not self.alive:
            raise SystemError("cluster is dead")

        self._remove_remote(self._client, self._prefix, self._log)
        self._remove_local(self._prefix, self._log)

        self._client = None
        self._scheduler = None
        self._workers = None
        self._network = None
        self._firewall = None

        self._log.info("Cluster %s destroyed.", self._prefix)

    @property
    def alive(self) -> bool:
        """
        Is cluster alive?
        """

        return self._scheduler is not None

    @property
    def dask_ipc(self) -> int:
        """
        Port used for Dask's interprocess communication
        """

        return self._dask_ipc

    @property
    def dask_dash(self) -> int:
        """
        Port used for Dask's dashboard
        """

        return self._dask_dash

    @property
    def dask_nanny(self) -> int:
        """
        Port used for Dask's nanny
        """

        return self._dask_nanny

    @property
    def scheduler(self) -> NodeABC:
        """
        A node running the Dask scheduler
        """

        if not self.alive:
            raise SystemError("cluster is dead")

        return self._scheduler

    @property
    def workers(self) -> List[NodeABC]:
        """
        A list of nodes running Dask workers
        """

        if not self.alive:
            raise SystemError("cluster is dead")

        return self._workers.copy()

    @property
    def prefix(self) -> str:
        """
        Name of cluster, used as a prefix in names of every component
        """

        return self._prefix

    @classmethod
    def _fn_private(cls, prefix: str) -> str:
        """
        Path to private key file
        """

        return os.path.join(os.getcwd(), f"{prefix:s}.key")

    @classmethod
    def _fn_public(cls, prefix: str) -> str:
        """
        Path to public key file
        """

        return f"{cls._fn_private(prefix):s}.pub"

    @classmethod
    def _remove_local(
        cls,
        prefix: str,
        log: Logger,
    ):

        if os.path.exists(cls._fn_private(prefix)):
            os.unlink(cls._fn_private(prefix))
        if os.path.exists(cls._fn_public(prefix)):
            os.unlink(cls._fn_public(prefix))

        for suffix in ("ca.key", "ca.crt", "node.key", "node.crt"):
            fn = os.path.join(os.getcwd(), f"{prefix:s}_{suffix:s}")
            if not os.path.exists(fn):
                continue
            log.info("Deleting local %s ...", fn)
            os.unlink(fn)

    @staticmethod
    def _remove_remote(
        client: Client,
        prefix: str,
        log: Logger,
    ):

        cats = [
            getattr(client, name)
            for name in (
                "servers",
                "networks",
                "ssh_keys",
                "firewalls",
            )
        ]

        for cat in cats:
            for item in cat.get_all():
                if not item.name.startswith(prefix):
                    log.warning("Not deleting %s ...", item.name)
                    continue
                log.info("Deleting remote %s ...", item.name)
                item.delete()

    @classmethod
    async def nuke(
        cls,
        prefix: str = PREFIX,
        tokenvar: str = TOKENVAR,
        log: Union[Logger, None] = None,
    ):
        """
        Nukes a possibly incomplete or broken cluster without prior connecting to it

        Args:
            prefix : Name of cluster, used as a prefix in names of every component.
            tokenvar : Name of the environment variable holding the cloud API login token.
            log : Allows to pass custom logger objects. Defaults to scherbelberg's own default logger.
        """

        log = getLogger(name=prefix) if log is None else log

        log.info("Creating cloud client ...")
        client = Client(token=os.environ[tokenvar])

        cls._remove_remote(client, prefix, log)
        cls._remove_local(prefix, log)

        log.info("Cluster %s nuked.", prefix)

    @classmethod
    async def from_new(
        cls,
        prefix: str = PREFIX,
        tokenvar: str = TOKENVAR,
        wait: float = WAIT,
        dask_ipc: int = DASK_IPC,
        dask_dash: int = DASK_DASH,
        dask_nanny: int = DASK_NANNY,
        scheduler: str = HETZNER_INSTANCE_TINY,
        worker: str = HETZNER_INSTANCE_TINY,
        image: str = HETZNER_IMAGE_UBUNTU,
        datacenter: str = HETZNER_DATACENTER,
        workers: int = WORKERS,
        log: Union[Logger, None] = None,
    ) -> ClusterABC:
        """
        Creates a new cluster

        Args:
            prefix : Name of cluster, used as a prefix in names of every component.
            tokenvar : Name of the environment variable holding the cloud API login token.
            wait : Timeout in seconds before actions are repeated or exceptions are raised.
            dask_ipc : Port used for Dask's interprocess communication.
            dask_dash : Port used for Dask's dashboard.
            dask_nanny : Port used for Dask's nanny.
            scheduler : Compute instance type used for Dask scheduler.
            worker : Compute instance type used for Dask workers.
            image : Operating system image.
            datacenter : Target data center.
            workers : Number of workers in cluster.
            log : Allows to pass custom logger objects. Defaults to scherbelberg's own default logger.
        Returns:
            Cluster object represeting an alive cluster.
        """

        log = getLogger(name=prefix) if log is None else log

        log.info("Creating cloud client ...")
        client = Client(token=os.environ[tokenvar])

        creator = await Creator.from_async(
            client=client,
            prefix=prefix,
            fn_public=cls._fn_public(prefix),
            fn_private=cls._fn_private(prefix),
            wait=wait,
            dask_ipc=dask_ipc,
            dask_dash=dask_dash,
            dask_nanny=dask_nanny,
            scheduler=scheduler,
            worker=worker,
            image=image,
            datacenter=datacenter,
            workers=workers,
            log=log,
        )

        return cls(
            client=client,
            scheduler=creator.scheduler,
            workers=creator.workers,
            network=creator.network,
            firewall=creator.firewall,
            dask_ipc=dask_ipc,
            dask_dash=dask_dash,
            dask_nanny=dask_nanny,
            prefix=prefix,
            wait=wait,
            log=log,
        )

    @classmethod
    async def from_existing(
        cls,
        prefix: str = PREFIX,
        tokenvar: str = TOKENVAR,
        wait: float = WAIT,
        log: Union[Logger, None] = None,
    ) -> ClusterABC:
        """
        Attaches to existing cluster

        Args:
            prefix : Name of cluster, used as a prefix in names of every component.
            tokenvar : Name of the environment variable holding the cloud API login token.
            wait : Timeout in seconds before actions are repeated or exceptions are raised.
            log : Allows to pass custom logger objects. Defaults to scherbelberg's own default logger.
        Returns:
            Cluster object represeting an alive cluster.
        """

        log = getLogger(name=prefix) if log is None else log

        log.info("Creating cloud client ...")
        client = Client(token=os.environ[tokenvar])

        log.info("Getting handle on scheduler ...")
        scheduler = await Node.from_name(
            name=f"{prefix:s}-node-scheduler",
            client=client,
            fn_private=cls._fn_private(prefix),
            prefix=prefix,
            wait=wait,
            log=log,
        )

        log.info("Getting handles on workers ...")
        workers = [
            Node(
                server=server,
                client=client,
                fn_private=cls._fn_private(prefix),
                prefix=prefix,
                wait=wait,
                log=log,
            )
            for server in client.servers.get_all()
            if server.name.startswith(prefix) and "-node-worker" in server.name
        ]

        log.info("Getting handle on firewall ...")
        firewall = client.firewalls.get_by_name(
            name=f"{prefix:s}-firewall",
        )

        log.info("Getting handle on network ...")
        network = client.networks.get_by_name(
            name=f"{prefix:s}-network",
        )

        log.info("Successfully attached to existing cluster.")
        return cls(
            client=client,
            scheduler=scheduler,
            workers=workers,
            network=network,
            firewall=firewall,
            dask_ipc=int(scheduler.labels["dask_ipc"]),
            dask_dash=int(scheduler.labels["dask_dash"]),
            dask_nanny=int(scheduler.labels["dask_nanny"]),
            prefix=prefix,
            wait=wait,
            log=log,
        )



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_core/creator.py: Creates a cluster

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from asyncio import create_task, gather, sleep
from logging import getLogger, Logger
import os
from typing import Dict, List, Union

from typeguard import typechecked

from hcloud import Client
from hcloud.datacenters.domain import Datacenter
from hcloud.firewalls.client import BoundFirewall
from hcloud.firewalls.domain import FirewallRule
from hcloud.images.domain import Image
from hcloud.networks.client import BoundNetwork
from hcloud.networks.domain import NetworkSubnet
from hcloud.servers.domain import Server
from hcloud.server_types.domain import ServerType
from hcloud.ssh_keys.client import BoundSSHKey

from .abc import CreatorABC, NodeABC
from .command import Command
from .const import (
    DASK_IPC,
    DASK_DASH,
    DASK_NANNY,
    WAIT,
    WORKERS,
    HETZNER_INSTANCE_TINY,
    HETZNER_IMAGE_UBUNTU,
    HETZNER_DATACENTER,
)
from .node import Node
from .ssl import create_ca, create_signed_cert, write_certs

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Creator(CreatorABC):
    """
    Creates a cluster of nodes.

    Mutable.
    """

    def __init__(
        self,
        client: Client,
        prefix: str,
        fn_public: str,
        fn_private: str,
        wait: float = WAIT,
        log: Union[Logger, None] = None,
    ):

        self._log = getLogger(name=prefix) if log is None else log

        self._client = client
        self._prefix = prefix
        self._fn_public = fn_public
        self._fn_private = fn_private
        self._wait = wait

        self._ssh_key = None
        self._firewall = None
        self._network = None
        self._scheduler = None
        self._workers = None

    async def create(
        self,
        dask_ipc: int = DASK_IPC,
        dask_dash: int = DASK_DASH,
        dask_nanny: int = DASK_NANNY,
        scheduler: str = HETZNER_INSTANCE_TINY,
        worker: str = HETZNER_INSTANCE_TINY,
        image: str = HETZNER_IMAGE_UBUNTU,
        datacenter: str = HETZNER_DATACENTER,
        workers: int = WORKERS,
    ):

        assert workers > 0

        assert dask_ipc >= 2 ** 10
        assert dask_dash >= 2 ** 10
        assert dask_nanny >= 2 ** 10

        assert len({dask_ipc, dask_dash, dask_nanny}) == 3

        await self._create_certs()
        self._ssh_key = await self._create_ssh_key()
        self._network = await self._create_network(ip_range="10.0.1.0/24")
        self._firewall = await self._create_firewall(dask_ipc, dask_dash, dask_nanny)

        self._log.info("Creating nodes ...")

        scheduler_task = create_task(
            self._create_node(
                suffix="scheduler",
                servertype=scheduler,
                datacenter=datacenter,
                image=image,
                ip="10.0.1.200",
                labels={
                    "dask_ipc": str(dask_ipc),
                    "dask_dash": str(dask_dash),
                    "dask_nanny": str(dask_nanny),
                },
            )
        )

        worker_tasks = [
            create_task(
                self._create_node(
                    suffix=f"worker{node:03d}",
                    servertype=worker,
                    datacenter=datacenter,
                    image=image,
                    ip=f"10.0.1.{100+node:d}",
                )
            )
            for node in range(workers)
        ]

        self._scheduler = await scheduler_task
        await self._scheduler.start_scheduler(dask_ipc=dask_ipc, dask_dash=dask_dash)

        self._workers = [await task for task in worker_tasks]
        await gather(
            *[
                worker.start_worker(
                    dask_ipc=dask_ipc,
                    dask_dash=dask_dash,
                    dask_nanny=dask_nanny,
                    scheduler_ip4=self._scheduler.public_ip4,
                )
                for worker in self._workers
            ]
        )

        self._log.info("Successfully created new cluster.")

    @property
    def scheduler(self) -> NodeABC:

        return self._scheduler

    @property
    def workers(self) -> List[NodeABC]:

        return self._workers.copy()

    @property
    def network(self) -> BoundNetwork:

        return self._network

    @property
    def firewall(self) -> BoundFirewall:

        return self._firewall

    async def _create_firewall(
        self,
        dask_ipc: int,
        dask_dash: int,
        dask_nanny: int,
    ) -> BoundFirewall:

        self._log.info("Creating firewall ...")

        _ = self._client.firewalls.create(
            name=f"{self._prefix:s}-firewall",
            rules=[
                FirewallRule(
                    direction="in",
                    protocol=protocol,
                    source_ips=["0.0.0.0/0", "::/0"],
                    destination_ips=[],
                    port=port,
                )
                for protocol, port in (
                    ("tcp", "22"),
                    ("icmp", None),
                    ("tcp", f"{dask_ipc:d}"),
                    ("tcp", f"{dask_dash:d}"),
                    ("tcp", f"{dask_nanny:d}"),
                )
            ],
        )

        self._log.info("Getting handle on firewall ...")

        return self._client.firewalls.get_by_name(
            name=f"{self._prefix:s}-firewall",
        )

    async def _create_network(self, ip_range: str) -> BoundNetwork:

        self._log.info("Creating network ...")

        _ = self._client.networks.create(
            name=f"{self._prefix:s}-network",
            ip_range=ip_range,
            subnets=[
                NetworkSubnet(
                    ip_range=ip_range,
                    type="cloud",
                    network_zone="eu-central",
                )
            ],
        )

        self._log.info("Getting handle on network ...")

        return self._client.networks.get_by_name(
            name=f"{self._prefix:s}-network",
        )

    async def _create_node(
        self,
        suffix: str,
        servertype: str,
        datacenter: str,
        image: str,
        ip: str,
        labels: Union[Dict[str, str], None] = None,
    ) -> NodeABC:

        name = f"{self._prefix:s}-node-{suffix:s}"

        self._log.info("Creating node %s ...", name)

        _ = self._client.servers.create(
            name=name,
            server_type=ServerType(name=servertype),
            image=Image(name=image),
            datacenter=Datacenter(name=datacenter),
            ssh_keys=[self._ssh_key],
            firewalls=[self._firewall],
            labels=labels,
        )

        self._log.info("Waiting for node %s to become available ...", name)

        while True:
            server = self._client.servers.get_by_name(name=name)
            if server.status == Server.STATUS_RUNNING:
                break
            await sleep(self._wait)

        self._log.info("Attaching network to node %s ...", name)

        server.attach_to_network(
            network=self._network,
            ip=ip,
        )

        self._log.info("Bootstrapping node %s ...", name)

        node = await Node.from_async(
            server=server,
            client=self._client,
            fn_private=self._fn_private,
            prefix=self._prefix,
            wait=self._wait,
            log=self._log,
        )

        await node.bootstrap()  # TODO param?
        await node.update()

        return node

    async def _create_ssh_key(self) -> BoundSSHKey:

        self._log.info("Creating ssh key ...")

        if os.path.exists(self._fn_private):
            raise SystemError("ssh private key file already exists")
        if os.path.exists(self._fn_public):
            raise SystemError("ssh public key file already exists")

        _, _ = await Command.from_list(
            [
                "ssh-keygen",
                "-f",
                self._fn_private,  # path to file
                "-P",
                "",  # no password
                "-t",
                "rsa",  # RSA
                "-b",
                "4096",  # bits for RSA
                "-C",
                f"{self._prefix:s}-key",  # comment
            ]
        ).run(wait=self._wait)

        self._log.info("Uploading ssh key ...")

        with open(self._fn_public, "r", encoding="utf-8") as f:
            public = f.read()

        _ = self._client.ssh_keys.create(
            name=f"{self._prefix:s}-key",
            public_key=public,
        )

        self._log.info("Getting handle on ssh key ...")

        return self._client.ssh_keys.get_by_name(
            name=f"{self._prefix:s}-key",
        )

    @typechecked
    async def _create_certs(self) -> None:

        self._log.info("Creating ssl certificates ...")

        ca_key, ca_cert = await create_ca(
            prefix=self._prefix,
        )
        await write_certs(
            ca_key,
            ca_cert,
            name=f"{self._prefix:s}_ca",
        )

        ca_key, ca_cert = await create_signed_cert(
            ca_key=ca_key,
            ca_cert=ca_cert,
            prefix=self._prefix,
        )
        await write_certs(
            ca_key,
            ca_cert,
            name=f"{self._prefix:s}_node",
        )

    @classmethod
    async def from_async(
        cls,
        client: Client,
        prefix: str,
        fn_public: str,
        fn_private: str,
        wait: float = WAIT,
        log: Union[Logger, None] = None,
        dask_ipc: int = DASK_IPC,
        dask_dash: int = DASK_DASH,
        dask_nanny: int = DASK_NANNY,
        scheduler: str = HETZNER_INSTANCE_TINY,
        worker: str = HETZNER_INSTANCE_TINY,
        image: str = HETZNER_IMAGE_UBUNTU,
        datacenter: str = HETZNER_DATACENTER,
        workers: int = WORKERS,
    ) -> CreatorABC:

        obj = cls(
            client=client,
            prefix=prefix,
            fn_public=fn_public,
            fn_private=fn_private,
            wait=wait,
            log=log,
        )

        await obj.create(
            dask_ipc=dask_ipc,
            dask_dash=dask_dash,
            dask_nanny=dask_nanny,
            scheduler=scheduler,
            worker=worker,
            image=image,
            datacenter=datacenter,
            workers=workers,
        )

        return obj



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_cli/ssh.py: ssh into cluster member

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from asyncio import run

import click
import os
import sys

from .._core.cluster import Cluster
from .._core.const import PREFIX, TOKENVAR, WAIT
from .._core.log import configure_log

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


async def _main(prefix, tokenvar, wait, hostname):

    cluster = await Cluster.from_existing(
        prefix=prefix,
        tokenvar=tokenvar,
        wait=wait,
    )

    nodes = {node.name.split("-node-")[1]: node for node in cluster.workers}
    nodes["scheduler"] = cluster.scheduler

    if hostname not in nodes.keys():
        print(
            f'"{hostname:s}" is unknown in cluster "{prefix:s}": '
            + ", ".join(nodes.keys())
        )
        sys.exit(1)

    dev_null = "\\\\.\\NUL" if sys.platform.startswith("win") else "/dev/null"

    host = await nodes[hostname].get_sshconfig(user=f"{prefix:s}user")
    cmd = [
        "ssh",
        "-o",
        "StrictHostKeyChecking=no",  # TODO security
        "-o",
        f"UserKnownHostsFile={dev_null:s}",  # TODO security
        "-o",
        "ConnectTimeout=5",
        "-o",
        "Compression=yes" if host.compression else "Compression=no",
        "-p",
        f"{host.port:d}",
        "-c",
        host.cipher,
        "-i",
        host.fn_private,
        "-q",
        f"{host.user:s}@{host.name:s}",
    ]
    os.execvpe(cmd[0], cmd, os.environ)


@click.command(short_help="ssh into cluster member")
@click.option("-p", "--prefix", default=PREFIX, type=str, show_default=True)
@click.option("-t", "--tokenvar", default=TOKENVAR, type=str, show_default=True)
@click.option("-a", "--wait", default=WAIT, type=float, show_default=True)
@click.argument("hostname", nargs=1, type=str)
def ssh(prefix, tokenvar, wait, hostname):

    configure_log()

    run(_main(prefix, tokenvar, wait, hostname))



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    src/scherbelberg/_core/ssl.py: Creating CAs and SSL keys

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from datetime import datetime, timedelta
from typing import Tuple

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
async def create_ca(
    prefix: str,
    valid_days: int = 365 * 2,
) -> Tuple[rsa.RSAPrivateKey, x509.Certificate]:

    assert valid_days > 0
    assert len(prefix) > 0

    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, "GL"),  # Greenland - why not?
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, f"{prefix:s} province"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, f"{prefix:s} locality"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, f"{prefix:s} organization"),
            x509.NameAttribute(NameOID.COMMON_NAME, f"{prefix:s} CA"),
        ]
    )
    issuer = subject  # identical

    key = rsa.generate_private_key(
        public_exponent=65537, key_size=4096, backend=default_backend()
    )

    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow() - timedelta(days=1))  # HACK
        .not_valid_after(datetime.utcnow() + timedelta(days=valid_days))
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True,  # ?
        )
        .sign(key, hashes.SHA256(), default_backend())  # self sign
    )

    return key, cert


@typechecked
async def create_signed_cert(
    ca_key: rsa.RSAPrivateKey,
    ca_cert: x509.Certificate,
    prefix: str,
    valid_days: int = 365 * 2,
) -> Tuple[rsa.RSAPrivateKey, x509.Certificate]:

    assert valid_days > 0
    assert len(prefix) > 0

    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, "GL"),  # Greenland - why not?
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, f"{prefix:s} province"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, f"{prefix:s} locality"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, f"{prefix:s} node"),
        ]
    )

    key = rsa.generate_private_key(
        public_exponent=65537, key_size=4096, backend=default_backend()
    )

    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(ca_cert.issuer)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow() - timedelta(days=1))  # HACK
        .not_valid_after(datetime.utcnow() + timedelta(days=valid_days))
        .add_extension(
            x509.SubjectAlternativeName(
                [x509.DNSName("scherbelberg.solarsystem")]
            ),  # Dask does not check the domain
            critical=False,
        )
        .sign(ca_key, hashes.SHA256(), default_backend())
    )

    return key, cert


@typechecked
async def write_certs(key: rsa.RSAPrivateKey, cert: x509.Certificate, name: str):

    assert len(name) > 0

    with open(f"{name:s}.key", "wb") as f:
        f.write(
            key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),  # ?
            )
        )

    with open(f"{name:s}.crt", "wb") as f:
        f.write(
            cert.public_bytes(
                encoding=serialization.Encoding.PEM,
            )
        )



# -*- coding: utf-8 -*-

"""

SCHERBELBERG
HPC cluster deployment and management for the Hetzner Cloud

https://github.com/pleiszenburg/scherbelberg

    setup.py: Used for package distribution

    Copyright (C) 2021-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the BSD 3-Clause License
("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://github.com/pleiszenburg/scherbelberg/blob/master/LICENSE
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import os

from setuptools import (
    find_packages,
    setup,
)

from docs.source.version import get_version

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# SETUP
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# List all versions of Python which are supported
python_minor_min = 8
python_minor_max = 10
confirmed_python_versions = [
    "Programming Language :: Python :: 3.{MINOR:d}".format(MINOR=minor)
    for minor in range(python_minor_min, python_minor_max + 1)
]

# Fetch readme file
with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    long_description = f.read()

# Define source directory (path)
SRC_DIR = "src"

# Version
version = get_version()

# Requirements
base_require = [
    "click",
    "dask",
    "hcloud",
    "pyyaml",
    "typeguard",
]
extras_require = {
    "base": base_require,
    "dev": [
        "black",
        "myst-parser",
        "python-lsp-server[all]",
        "setuptools",
        "sphinx",
        "sphinx-click",
        "sphinx_rtd_theme",
        "sphinx-autodoc-typehints",
        "twine",
        "wheel",
    ],
}
extras_require["all"] = list(
    {rq for target in extras_require.keys() for rq in extras_require[target]}
)

# Install package
setup(
    name="scherbelberg",
    packages=find_packages(SRC_DIR),
    package_dir={"": SRC_DIR},
    version=version,
    description="HPC cluster deployment and management for the Hetzner Cloud",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Sebastian M. Ernst",
    author_email="ernst@pleiszenburg.de",
    url="https://github.com/pleiszenburg/scherbelberg",
    download_url=f"https://github.com/pleiszenburg/scherbelberg/archive/v{version:s}.tar.gz",
    license="BSD",
    keywords=["HPC", "cluster", "dask"],
    scripts=[],
    include_package_data=True,
    python_requires=">=3.{MINOR:d}".format(MINOR=python_minor_min),
    setup_requires=[],
    install_requires=base_require,
    extras_require=extras_require,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "scherbelberg = scherbelberg._cli:cli",
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
    ]
    + confirmed_python_versions
    + [
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: System :: Clustering",
        "Topic :: Scientific/Engineering",
        "Topic :: System",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
    ],
)
